# Copyright (c) 2012-2014, GPy authors (see AUTHORS.txt).
# Licensed under the BSD 3-clause license (see LICENSE.txt)
"""
Examples for GPy.

The examples in this package usually depend on `pods <https://github.com/sods/ods>`_ so make sure 
you have that installed before running examples.
"""
from . import classification
from . import regression
from . import dimensionality_reduction
from . import non_gaussian
